/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
* Module Name:
*   lec_data.c
*   
* Abstract:
*   This file implements the LEC_DATA module which handles reception and
*   transmission of data over Emulated LANs.
*                                                                             
* Authors:
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date       Name  Description 
*   27-Jan-95  TLR   - Created.
*
* Description:
*   
*/

#ifdef _DEBUG
#define STATIC
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#else
#define STATIC static
#define assert(expr)
#endif

/* Include general headers. */
#include "g_types.h"
#include "codes.h"
#include "atm.h"
#include "utl.h"
#include "utl_os.h"
#include "system.h"
#include "af_lane.h"
#include "g_endian.h"
#include "g_event.h"

/* Include headers for this module's interfaces. */
#include "lec_mgt.h"
#include "lec.h"

/* Include headers for the this module's neighbors. */
#include "svc_info.h"
#include "cm_sap.h"
#include "lec_ctrl.h"
#include "map.h"


/****************************************************************************
 * Internal Data Structures
 ****************************************************************************/

/* MCAST_ENTRY
 *  This structure describes a single multicast address that is stored in the
 *  multicast address list of a LEC.
 *
 *  list_hdr....This header is provided by the Map module.  It allows this
 *              structure to be stored in a map_list.
 *
 *  addr........This is the multicast address.  It is placed directly after the
 *              list header so it can be used as the search key of the list.
 *
 *  count.......The instance count for this address.  This value is equal to
 *              the number of times this address has been added to the
 *              multicast address list.
 */
typedef struct
   {
   MAP_LIST_HDR   list_hdr;
   ADDR_48        addr;
   UINT16         count;
   } MCAST_ENTRY;


/* LD_ELAN_CONTEXT
 *  Context block for a single ELAN membership.  An instance of this structure
 *  is created each time the LEC_DATA module registers a new ELAN membership.
 *
 *  p_next.................Next pointer that allows this structure to be
 *                         stored in a simple list.
 *
 *  lec_handle.............The handle of this LEC module instance.  This is
 *                         the handle that was generated by the lec_create
 *                         function.
 *
 *  client_context.........The context supplied by the ELAN client during
 *                         lec registration.  There is one context for each
 *                         ELAN.
 *
 *  os_handle..............OS handle for this physical port.
 *
 *  cm_handle..............Connection Manager handle for this physical port.
 *
 *  lc_elan_handle.........Handle for the ELAN instance of the LEC_CTRL module.
 *
 *  sap_handle.............Handle of the registered DATA SAP.
 *
 *  lan_type...............LAN Type of the joined (or joining) ELAN.
 *
 *  direct_conn_context....Connection Context for data direct VCCs.
 *
 *  bus_conn_context.......Connection Context for BUS VCCs.
 *
 *  mcast_list.............Handle for the multicast address list.  This list is
 *                         managed by the mapping module.
 *
 *  elan_up................This flag indicates whether or not the ELAN is fully
 *                         joined and ready for network traffic.
 *
 *  lec_id.................LEC_ID of this LEC in the ELAN.  Used in the LAN
 *                         emulation header for data packets.
 *
 *  mcast_enb..............Multicast enable flag.
 *
 *  bcast_enb..............Briadcast enable flag.
 *
 *  prom_enb...............Promiscuous mode enable flag.
 *
 *  lec_event_callback.....Callback for ELAN events.
 *
 *  lec_rcv_callback.......Callback for received packets.
 *
 *  lec_xmt_done_callback..Callback for transmit done indications.
 */
typedef struct _ld_elan
   {
   struct _ld_elan        *p_next;
   char                   *p_text;
   UINT32                  event_mask;
   HANDLE                  lec_handle;
   HANDLE                  client_context;
   HANDLE                  os_handle;
   HANDLE                  cm_handle;
   HANDLE                  lc_elan_handle;
   HANDLE                  sap_handle;
   LAN_TYPE                lan_type;
   LEC_CONN_CONTEXT        direct_conn_context;
   LEC_CONN_CONTEXT        mcast_conn_context;
   LEC_CONN_CONTEXT        ctrl_conn_context;
   HANDLE                  mcast_list;
   BOOLEAN                 elan_up;
   UINT16                  lec_id;
   BOOLEAN                 mcast_enb;
   BOOLEAN                 bcast_enb;
   BOOLEAN                 prom_enb;
   LEC_EVENT_CALLBACK      lec_event_callback;
   LEC_RCV_CALLBACK        lec_rcv_callback;
   LEC_XMT_DONE_CALLBACK   lec_xmt_done_callback;
   UINT32                  bus_frames_rcvd;
   UINT32                  direct_frames_rcvd;
   UINT32                  rcv_frames_discarded;
   UINT32                  echos_suppressed;
   } LD_ELAN_CONTEXT;

#define EMOD   MOD_LEC_DATA
#define EINST  p_elan->p_text
#define EMASK  p_elan->event_mask

/*
 * Declare a simple list type for ELAN Contexts.
 */
SIMPLE_LIST (LD_ELAN_CONTEXT, ELAN_LIST);

/* LD_CONTEXT
 *  Context structure for LEC instances.  There is one such instance for each
 *  physical port that provides LAN Emulation Client capabilities.
 *
 *  os_handle..OS Handle for this physical port.
 *
 *  cm_handle..Connection Manager handle for this physical port.
 *
 *  lc_handle..LEC_CTRL handle for this physical port.
 *
 *  elan_list..List of ELAN contexts that are registered to this physical port.
 */
typedef struct
   {
   HANDLE      os_handle;
   HANDLE      cm_handle;
   HANDLE      lc_handle;
   HANDLE      addr_reg_handle;
   ELAN_LIST   elan_list;
   } LD_CONTEXT;


/****************************************************************************
 * Internal Routines
 ****************************************************************************/

/*++
* ========================
* = elan_status_callback =
* ========================
*
* Overview:
*
* Arguments:
*   elan_handle      - (IN) Handle returned by lec_register.
*   elan_available   - (IN) ELAN Availability status (boolean).
*   mac_addr         - (IN) 48-bit address of this client.
*   lan_mtu          - (IN) LAN MTU negotiated during the join.
*   p_elan_name      - (IN) Name of ELAN that was joined (if event is
*                           "NETWORK_AVAILABLE").
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The elan_up state variable takes on the value of elan_available.
--*/
static void elan_status_callback (HANDLE   elan_handle,
                                  BOOLEAN  elan_available,
                                  LAN_MTU  lan_mtu,
                                  char    *p_elan_name,
                                  UINT16   lec_id)
   {
   LD_ELAN_CONTEXT   *p_elan;
   LEC_EVENT          lec_event;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   p_elan->elan_up = elan_available;

   lec_event = elan_available ? LEC_NETWORK_AVAILABLE :
                                LEC_NETWORK_UNAVAILABLE;

   EVENT (EM_EVENT, ("ELAN Status Notification - ELAN is %s\n",
                     elan_available ? "Up" : "Down"));

   p_elan->lec_event_callback (p_elan->client_context,
                               lec_event,
                               lan_mtu,
                               p_elan_name);

   p_elan->lec_id = lec_id;
   }


/*++
* ========================
* = ld_xmt_done_callback =
* ========================
*
* Overview:
*
* Arguments:
*   conn_context
*   p_packet
*   user_data
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The elan_up state variable takes on the value of elan_available.
--*/
STATIC void ld_xmt_done_callback (HANDLE    conn_context,
                                  void     *p_packet,
                                  UINT32    user_data)
   {
   LEC_CONN_CONTEXT   *p_conn;
   LD_ELAN_CONTEXT    *p_elan;

   p_conn = (LEC_CONN_CONTEXT *) conn_context;
   p_elan = (LD_ELAN_CONTEXT  *) p_conn->ld_elan_handle;

   if (user_data & USER_DATA_INTERNAL)
      {
      EVENT (EM_ASSERT, ("Transmit-Done Indication on Internal Data\n"));
      assert (FALSE);
      }

   p_elan->lec_xmt_done_callback (p_elan->client_context,
                                  p_packet);

   return;
   }


/*++
* ===================
* = ld_rcv_callback =
* ===================
*
* Overview:
*
* Arguments:
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The elan_up state variable takes on the value of elan_available.
--*/
static void ld_rcv_callback (HANDLE     conn_context,
                             HANDLE     conn_handle,
                             UINT32     length,
                             UINT32     user_data,
                             AAL_DATA  *p_aal_data,
                             void     **pp_packet)
   {
   LEC_CONN_CONTEXT  *p_conn_context;
   LD_ELAN_CONTEXT   *p_elan;
   UINT16             lec_hdr;
   HDR_802_3          hdr_802_3;
   HDR_802_5          hdr_802_5;
   UINT8             *p_dst_addr;
   HANDLE             not_used;
   STATUS             status;
   UINT32             bytes_copied;

   p_conn_context = (LEC_CONN_CONTEXT *) conn_context;
   p_elan         = (LD_ELAN_CONTEXT *)  p_conn_context->ld_elan_handle;

   /* If this packet was received on a control VCC, shunt it off to the
    * LEC_CTRL module's receive handler.
    */
   if (p_conn_context->conn_type == LEC_CTRL)
      {
      lc_sap_rcv (conn_context,
                  conn_handle,
                  length,
                  user_data,
                  p_aal_data,
                  pp_packet);
      return;
      }

   /* Extract the LAN Emulation header from the packet. */

   os_buff_hdr_get (*pp_packet, &lec_hdr, &bytes_copied);
   if (bytes_copied != 2)
      {
      EVENT (EM_NERR, ("Unable to extract LE Header from received frame\n"));
      return;
      }

   lec_hdr = ntoh16 (lec_hdr);

   /* If the LE header is equal to this LEC's lec_id, discard the packet
    * because it was sent by this LEC.
    */
   if (lec_hdr == p_elan->lec_id)
      {
      EVENT (EM_RDATA, ("Echo Frame Suppressed\n"));
      p_elan->echos_suppressed += 1;
      return;
      }

   /* Check to see if the received packet is a LAN Emulation control packet.
    * If it is, it must be sent over to the LEC_CTRL module for further
    * processing.
    *
    * Note that the conditional tests for >= the marker value.  This
    * causes all non-data frames to be sent on the slow path to the LEC_CTRL
    * module.  It will further check to ensure the header contains a valid
    * number.
    */
   if (lec_hdr >= LE_CTRL_MARKER)
      {
      lc_sap_rcv (conn_context,
                  conn_handle,
                  length,
                  user_data,
                  p_aal_data,
                  pp_packet);
      return;
      }

   /* If the event_mask has the "log received data frames" bit set, copy the
    * frame into a buffer and log its contents.
    */
   if (p_elan->event_mask & EM_RDATA)
      {
      char   buffer[128];
      char   out[50];
      UINT32 i, j;

      os_buff_hdr_copy (*pp_packet,
                        length,
                        buffer,
                        128,
                        &bytes_copied);

      EVENT (EM_RDATA, ("Data Frame Received, length = %d\n", length));
      j = 0;
      for (i = 0; i < bytes_copied; i++)
         {
         sprintf (&out[j], "%02X ", ((UINT16) buffer[i]) & 0xff);
         j += 3;
         if ((i % 16) == 15)
            {
            j = 0;
            EVENT (EM_RDATA, ("%s\n", out));
            }
         }
      if ((i % 16) != 0)
         EVENT (EM_RDATA, ("%s\n", out));
      }

   if (p_conn_context->conn_type == LEC_DIRECT)
      p_elan->direct_frames_rcvd +=1;

   if (p_conn_context->conn_type == LEC_MCAST)
      p_elan->bus_frames_rcvd += 1;

   /* If the packet arrived on a data-direct VCC, or promiscuous mode is
    * enabled, forward it to the upper layer without further ado.
    */
   if ((p_conn_context->conn_type == LEC_DIRECT) ||
       (p_elan->prom_enb == TRUE))
      {
      p_elan->lec_rcv_callback (p_elan->client_context, length, pp_packet);
      return;
      }

   /* Handle BUS traffic according to the filters. */

   /* Extract the destination MAC address from the received packet.  This
    * has to be done differently for different LAN types.
    */
   switch (p_elan->lan_type)
      {
      case LAN_802_3 :
         os_buff_hdr_copy (*pp_packet,
                           length,
                           &hdr_802_3,
                           sizeof (HDR_802_3),
                           &bytes_copied);
         if (bytes_copied != sizeof (HDR_802_3))
            {
            p_elan->rcv_frames_discarded += 1;
            return;
            }
         p_dst_addr = hdr_802_3.dst_addr;
         break;
      case LAN_802_5 :
         os_buff_hdr_copy (*pp_packet,
                          length,
                          &hdr_802_5,
                          sizeof (HDR_802_5),
                          &bytes_copied);
         if (bytes_copied != sizeof (HDR_802_5))
            {
            p_elan->rcv_frames_discarded += 1;
            return;
            }
         p_dst_addr = hdr_802_5.dst_addr;
         break;
      default :
         assert (FALSE);
         return;
      }

   /* If the received packet is a unicast packet and its destination address
    * is a registered address of this LEC, send the packet up.
    */
   if (lc_dest_is_registered (p_elan->lc_elan_handle, p_dst_addr))
      {
      p_elan->lec_rcv_callback (p_elan->client_context, length, pp_packet);
      return;
      }

   /* If broadcasts are enabled and the received packet was sent to a
    * broadcast address, send the packet up.
    */
   if ((p_elan->bcast_enb == TRUE) &&
       (ESI_IS_BCAST (p_dst_addr)))
      {
      p_elan->lec_rcv_callback (p_elan->client_context, length, pp_packet);
      return;
      }

   if (ESI_IS_MCAST (p_dst_addr))
      {
      /* If the packet is a multicast packet and multicasts are enabled,
       * send the packet up.
       */
      if (p_elan->mcast_enb == TRUE)
         {
         p_elan->lec_rcv_callback (p_elan->client_context, length, pp_packet);
         return;
         }

      /* If multicasts are not enabled, search the multicast address list
       * to see if this address is to be received.  If the address is found,
       * send the packet up to the upper layer.
       */
      status = map_key_search (p_elan->mcast_list, p_dst_addr, &not_used);
      if (status == STATUS_K_SUCCESS)
         {
         p_elan->lec_rcv_callback (p_elan->client_context, length, pp_packet);
         return;
         }
      }

   /* Discard any packet that does not meet the above criteria. */

   p_elan->rcv_frames_discarded += 1;
   return;
   }


STATIC BOOLEAN mcast_element_delete (HANDLE mcast_handle,
                                     HANDLE os_handle)
   {
   MCAST_ENTRY *p_entry;

   p_entry = (MCAST_ENTRY *) mcast_handle;

   os_mem_dealloc (os_handle, p_entry);

   return TRUE;
   }

/****************************************************************************
 * LEC Interface Functions
 ****************************************************************************/

/*++
* ==============
* = lec_create =
* ==============
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS lec_create (HANDLE   os_handle,
                   HANDLE   cm_handle,
                   HANDLE   line_up_handle,
                   char    *p_text,
                   HANDLE  *p_lec_handle)
   {
   LD_CONTEXT  *p_context;
   STATUS       status;

   /* Allocate storage for the data context.  Return failure status if memory
    * could not be allocated.
    */
   p_context = (LD_CONTEXT *) os_mem_alloc (os_handle, sizeof (LD_CONTEXT));
   if (p_context == NULL)
      return STATUS_K_RESOURCES;

   /* Initialize the context's neighbor handles. */

   p_context->os_handle       = os_handle;
   p_context->cm_handle       = cm_handle;

   /* Initialize the list of ELANs. */

   utl_list_init (p_context->elan_list);

   /* Create an instance of the LEC_CTRL module.  If there is an error, back
    * out of our previous operations and return failure status.
    */
   status = lc_create (os_handle,
                       cm_handle,
                       line_up_handle,
                       elan_status_callback,
                       &p_context->lc_handle);
   if (status != STATUS_K_SUCCESS)
      {
      os_mem_dealloc (os_handle, p_context);
      return STATUS_K_RESOURCES;
      }

   /* Return the lec handle with success status. */

   *p_lec_handle = (HANDLE) p_context;
   return STATUS_K_SUCCESS;
   }

/*++
* ===============
* = lec_destroy =
* ===============
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_destroy (HANDLE  lec_handle)
   {
   LD_CONTEXT        *p_context;
   LD_ELAN_CONTEXT   *p_elan;

   if (lec_handle == NULL)
      return;

   /* Get a pointer to the LEC context data block. */

   p_context = (LD_CONTEXT *) lec_handle;

   /* Deregister each ELAN in the elan list. */

   p_elan = p_context->elan_list.p_root;
   while (p_elan != NULL)
      {
      lec_deregister ((HANDLE) p_elan);
      p_elan = p_context->elan_list.p_root;
      }

   /* Request the destruction of the LEC_CTRL module instance. */

   lc_destroy (p_context->lc_handle);

   /* Deallocate the memory for the LEC context data block. */

   os_mem_dealloc (p_context->os_handle, p_context);
   }

/*++
* ================
* = lec_register =
* ================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS lec_register (HANDLE                 lec_handle,
                     HANDLE                 context,
                     HANDLE                 addr_reg_handle,
                     LEC_EVENT_CALLBACK     event_callback,
                     LEC_RCV_CALLBACK       rcv_callback,
                     LEC_XMT_DONE_CALLBACK  xmt_done_callback,
                     ESI                    mac_addr,
                     LAN_TYPE               lan_type,
                     LAN_MTU                max_frame_size,
                     char                  *p_elan_name,
                     INIT_METHOD            init_method,
                     ADDR_ATM               manual_atm_addr,
                     BOOLEAN                proxy_flag,
                     HANDLE                 lport_handle,
                     char                  *p_text,
                     HANDLE                *p_elan_handle)
   {
   LD_CONTEXT        *p_context;
   LD_ELAN_CONTEXT   *p_elan;
   STATUS             status;

   /* Obtain a pointer to the module context. */

   p_context = (LD_CONTEXT *) lec_handle;

   /* Allocate memory for a new ELAN context.  If the allocation fails, return
    * error status.
    */
   p_elan = (LD_ELAN_CONTEXT *) os_mem_alloc (p_context->os_handle,
                                              sizeof (LD_ELAN_CONTEXT));
   if (p_elan == NULL)
      return STATUS_K_RESOURCES;

   p_elan->event_mask = EM_DEFAULT;
   p_elan->p_text     = p_text;

   /* Initialize the ELAN context's neighbor handles. */

   p_elan->lec_handle     = lec_handle;
   p_elan->client_context = context;
   p_elan->os_handle      = p_context->os_handle;
   p_elan->cm_handle      = p_context->cm_handle;
   p_elan->lan_type       = lan_type;

   p_elan->bus_frames_rcvd       = 0;
   p_elan->direct_frames_rcvd    = 0;
   p_elan->rcv_frames_discarded  = 0;
   p_elan->echos_suppressed      = 0;

   /* Store the callback addresses in the ELAN context. */

   p_elan->lec_event_callback    = event_callback;
   p_elan->lec_rcv_callback      = rcv_callback;
   p_elan->lec_xmt_done_callback = xmt_done_callback;

   /* Set the "up" state to false since the ELAN has not been joined yet. */

   p_elan->elan_up   = FALSE;

   /* Initialize the multicast, broadcast, and promiscuous enables. */

   p_elan->mcast_enb = FALSE;
   p_elan->bcast_enb = FALSE;
   p_elan->prom_enb  = FALSE;

   /* Allocate the multicast address list.  If there is an error, back out of
    * our operations thus far and return error status.
    */
   status = map_list_alloc (p_context->os_handle,
                            sizeof (ADDR_48),
                            &p_elan->mcast_list);
   if (status != STATUS_K_SUCCESS)
      {
      EVENT (EM_SERR, ("Failed to allocate multicast list.\n"));
      os_mem_dealloc (p_context->os_handle, p_elan);
      return STATUS_K_RESOURCES;
      }

   /* Register with the LEC_CTRL module to cause it to allocate an
    * ELAN instance.  If there is an error, back out of our operations so
    * far and return error status.
    */
   status = lc_register (p_context->lc_handle,
                         (HANDLE) p_elan,
                         addr_reg_handle,
                         mac_addr,
                         lan_type,
                         max_frame_size,
                         p_elan_name,
                         init_method,
                         manual_atm_addr,
                         proxy_flag,
                         lport_handle,
                         p_text,
                         &p_elan->lc_elan_handle);
   if (status != STATUS_K_SUCCESS)
      {
      EVENT (EM_SERR, ("Failed to register lec_ctrl instance.\n"));
      map_list_dealloc (p_elan->mcast_list,
                        mcast_element_delete,
                        p_elan->os_handle);
      os_mem_dealloc   (p_context->os_handle, p_elan);
      return STATUS_K_RESOURCES;
      }

   /* Pre-initialize the data-direct and BUS connection context blocks. */

   p_elan->direct_conn_context.ld_elan_handle   = (HANDLE) p_elan;
   p_elan->direct_conn_context.lc_elan_handle   = p_elan->lc_elan_handle;
   p_elan->direct_conn_context.conn_type        = LEC_DIRECT;

   p_elan->mcast_conn_context.ld_elan_handle    = (HANDLE) p_elan;
   p_elan->mcast_conn_context.lc_elan_handle    = p_elan->lc_elan_handle;
   p_elan->mcast_conn_context.conn_type         = LEC_MCAST;

   p_elan->ctrl_conn_context.ld_elan_handle     = (HANDLE) p_elan;
   p_elan->ctrl_conn_context.lc_elan_handle     = p_elan->lc_elan_handle;
   p_elan->ctrl_conn_context.conn_type          = LEC_CTRL;

   /* Add the ELAN context to the ELAN list in the module context. */

   utl_list_add (p_context->elan_list, p_elan);

   /* Register with the CM as a SAP.  This is the Data-SAP for this ELAN.
    * Again, back out if there is an error.
    */
   status = cm_sap_register (p_elan->cm_handle,
                             p_elan->lc_elan_handle,
                             ld_rcv_callback,
                             lc_sap_vc_notify,
                             lc_sap_connect,
                             lc_sap_link_status,
                             ld_xmt_done_callback,
                             "LAN Emulation Convergence",
                             &p_elan->sap_handle);
   if (status != STATUS_K_SUCCESS)
      {
      EVENT (EM_SERR, ("Failed to register as an NSAP client.\n"));
      lc_deregister    (p_elan->lc_elan_handle);
      map_list_dealloc (p_elan->mcast_list,
                        mcast_element_delete,
                        p_elan->os_handle);
      os_mem_dealloc   (p_context->os_handle, p_elan);
      return STATUS_K_RESOURCES;
      }

   /* Registration is complete.  Notify the LEC_CTRL module that it may begin
    * the ELAN join.
    */
   lc_join_start (p_elan->lc_elan_handle,
                  p_elan->sap_handle,
                  (HANDLE) &p_elan->direct_conn_context,
                  (HANDLE) &p_elan->mcast_conn_context,
                  (HANDLE) &p_elan->ctrl_conn_context);

   /* Return the new handle with success status. */

   *p_elan_handle = (HANDLE) p_elan;
   return STATUS_K_SUCCESS;
   }

/*++
* ==================
* = lec_deregister =
* ==================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_deregister (HANDLE elan_handle)
   {
   LD_CONTEXT        *p_context;
   LD_ELAN_CONTEXT   *p_elan;

   /* Get a pointer to the ELAN context data block. */

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   EVENT (EM_EVENT, ("ELAN Deregistered\n"));

   /* Remove this data block from the LEC instance's list of ELANs. */

   p_context = (LD_CONTEXT *) p_elan->lec_handle;
   utl_list_delete (p_context->elan_list, p_elan);

   /* Deallocate the multicast address list.  This uses the callback
    * "mcast_element_delete" to destroy each list element.
    */
   map_list_dealloc (p_elan->mcast_list,
                     mcast_element_delete,
                     p_elan->os_handle);

   /* Deregister the LEC_CTRL instance for this ELAN.  This will in turn
    * deregister the le-arp instance.
    */
   lc_deregister (p_elan->lc_elan_handle);

   /* Deregister with the Connection Manager. */

   cm_sap_unregister (p_elan->sap_handle);

   /* Deallocate the ELAN context data block. */

   os_mem_dealloc (p_elan->os_handle, p_elan);
   }


/*++
* ======================
* = lec_event_mask_set =
* ======================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_event_mask_set (HANDLE elan_handle,
                         UINT32 event_mask)
   {
   LD_ELAN_CONTEXT   *p_elan;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   p_elan->event_mask = event_mask;
   lc_event_mask_set (p_elan->lc_elan_handle, event_mask);
   }


/*++
* ====================
* = lec_mac_addr_add =
* ====================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS lec_mac_addr_add (HANDLE elan_handle,
                         ESI    mac_addr)
   {
   LD_ELAN_CONTEXT   *p_elan;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   return lc_mac_addr_add (p_elan->lc_elan_handle, mac_addr);
   }


/*++
* ====================
* = lec_mac_addr_del =
* ====================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_mac_addr_del (HANDLE elan_handle,
                       ESI    mac_addr)
   {
   LD_ELAN_CONTEXT   *p_elan;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   lc_mac_addr_del (p_elan->lc_elan_handle, mac_addr);
   }


/*++
* ===================
* = lec_filters_set =
* ===================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_filters_set (HANDLE  elan_handle,
                      BOOLEAN mcast_enb,
                      BOOLEAN bcast_enb,
                      BOOLEAN prom_enb)
   {
   LD_ELAN_CONTEXT   *p_elan;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   p_elan->mcast_enb = mcast_enb;
   p_elan->bcast_enb = bcast_enb;
   p_elan->prom_enb  = prom_enb;

   EVENT (EM_EVENT, ("Receive Packet Filters modified\n"));

   return;
   }

/*++
* ===================
* = lec_filters_get =
* ===================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_filters_get (HANDLE     elan_handle,
                      BOOLEAN   *p_mcast_enb,
                      BOOLEAN   *p_bcast_enb,
                      BOOLEAN   *p_prom_enb)
   {
   LD_ELAN_CONTEXT   *p_elan;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   *p_mcast_enb = p_elan->mcast_enb;
   *p_bcast_enb = p_elan->bcast_enb;
   *p_prom_enb  = p_elan->prom_enb;

   return;
   }

/*++
* =================
* = lec_mcast_add =
* =================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS lec_mcast_add (HANDLE  elan_handle,
                      ESI     addr)
   {
   LD_ELAN_CONTEXT   *p_elan;
   STATUS             status;
   HANDLE             mcast_handle;
   MCAST_ENTRY       *p_mcast_entry;

   /* Obtain a pointer to the ELAN instance structure. */

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   /* Search the multicast address list for an entry containing the
    * desired MAC address.
    */
   status = map_key_search (p_elan->mcast_list, &addr, &mcast_handle);
   if (status == STATUS_K_SUCCESS)
      {
      /* If a matching entry is found, simply increment the instance count
       * of that entry.  There is no need to create an additional entry.
       */
      p_mcast_entry = (MCAST_ENTRY *) mcast_handle;
      p_mcast_entry->count += 1;
      EVENT (EM_EVENT, ("Mcast multiple added - %02x-%02x-%02x-%02x-%02x-%02x\n",
                        addr[0], addr[1],addr[2], addr[3],addr[4], addr[5]));
      return STATUS_K_SUCCESS;
      }
   else
      {
      /* If no matching entry is found, a new entry must be created with
       * an instance count of 1.  This entry is placed in the multicast
       * address list.
       */
      p_mcast_entry = (MCAST_ENTRY *) os_mem_alloc (p_elan->os_handle,
                                                    sizeof (MCAST_ENTRY));
      if (p_mcast_entry == NULL)
         {
         EVENT (EM_SERR, ("No room for mcast esi in filter\n"));
         return STATUS_K_RESOURCES;
         }
      
      ESI_COPY (addr, p_mcast_entry->addr);
      p_mcast_entry->count     = 1;
      map_element_insert (p_elan->mcast_list, (HANDLE) p_mcast_entry);
      EVENT (EM_EVENT, ("Mcast added - %02x-%02x-%02x-%02x-%02x-%02x\n",
                        addr[0], addr[1],addr[2], addr[3],addr[4], addr[5]));
      }

   return STATUS_K_SUCCESS;
   }

/*++
* ====================
* = lec_mcast_delete =
* ====================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS lec_mcast_delete (HANDLE  elan_handle,
                         ESI     addr)
   {
   LD_ELAN_CONTEXT   *p_elan;
   STATUS             status;
   HANDLE             mcast_handle;
   MCAST_ENTRY       *p_mcast_entry;

   /* Obtain a pointer to the ELAN instance structure. */

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   /* Search for an entry in the multicast address list that contains the
    * specified MAC address.
    */
   status = map_key_search (p_elan->mcast_list, &addr, &mcast_handle);
   if (status == STATUS_K_SUCCESS)
      {
      /* If a matching entry is found, decrement its instance count. */

      p_mcast_entry = (MCAST_ENTRY *) mcast_handle;
      p_mcast_entry->count -= 1;

      EVENT (EM_EVENT, ("Mcast deleted - %02x-%02x-%02x-%02x-%02x-%02x\n",
                        addr[0], addr[1],addr[2], addr[3],addr[4], addr[5]));

      /* If the instance count is now zero, the entry must be deleted from
       * the multicast address list.
       */
      if (p_mcast_entry->count == 0)
         {
         map_element_delete (p_elan->mcast_list, mcast_handle);
         os_mem_dealloc (p_elan->os_handle, p_mcast_entry);
         }

      return STATUS_K_SUCCESS;
      }
   else
      {
      /* If no matching entry was found, return error status. */

      EVENT (EM_NERR, ("Mcast not found for delete - "
                       "%02x-%02x-%02x-%02x-%02x-%02x\n",
                       addr[0], addr[1],addr[2], addr[3],addr[4], addr[5]));

      return STATUS_K_NOT_FOUND;
      }
   }

/*++
* ==================
* = lec_mcast_load =
* ==================
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void lec_mcast_load (HANDLE   elan_handle,
                     UINT32   table_size,
                     ADDR_48  addr_array[])
   {
   LD_ELAN_CONTEXT   *p_elan;
   MCAST_ENTRY       *p_mcast_entry;
   UINT16             index;

   /* Obtain a pointer to the ELAN instance structure. */

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   /* The multicast address list must first be cleared. */

   map_list_traverse (p_elan->mcast_list,
                      mcast_element_delete,
                      p_elan->os_handle);

   /* For each entry in the address array, create a multicast list element
    * with an instance count of 1.
    */
   for (index = 0; index < table_size; index++)
      {
      p_mcast_entry = (MCAST_ENTRY *) os_mem_alloc (p_elan->os_handle,
                                                    sizeof (MCAST_ENTRY));
      if (p_mcast_entry != NULL)
         {
         ESI_COPY (addr_array[index], p_mcast_entry->addr);
         p_mcast_entry->count     = 1;
         map_element_insert (p_elan->mcast_list, (HANDLE) p_mcast_entry);
         }
      }

   EVENT (EM_EVENT, ("Mcast table loaded.\n"));
   return;
   }

typedef struct
   {
   ADDR_48  *addr_array;
   UINT32   *count_array;
   UINT32    index;
   UINT32    max;
   } ARRAY_BLOCK;

STATIC BOOLEAN mcast_get_handler (HANDLE entry_handle,
                                  HANDLE param)
   {
   ARRAY_BLOCK *p_array_block;
   MCAST_ENTRY *p_mcast_entry;

   p_mcast_entry = (MCAST_ENTRY *) entry_handle;
   p_array_block = (ARRAY_BLOCK *) param;

   if (p_array_block->index >= p_array_block->max)
      return FALSE;

   ESI_COPY (p_mcast_entry->addr,
             p_array_block->addr_array[p_array_block->index]);
   p_array_block->count_array[p_array_block->index] = p_mcast_entry->count;

   p_array_block->index += 1;

   return TRUE;
   }


/*++
* =================
* = lec_mcast_get =
* =================
*
* Description:
*   The target arrays (passed in as parameters) must be filled with data from
*   the multicast address list.  Since the address list is stored in a map
*   list, the list must be traversed using the map_list_traverse method.  This
*   method calls a handler for each element in the list as it is traversed.
*   The handler must have access to the target arrays and the array index to
*   properly fill them.
*
*   This routine allocates a structure that contains pointers to the target
*   arrays, an array index, and a maximum value.  The address of this structure
*   is passed into the traversal method to be passed again onto the handler
*   each time it is called.  The handler places each element into the arrays
*   at the index point and increments the index.
*
* Testing:
*
* Issues:
*
--*/
void lec_mcast_get (HANDLE   elan_handle,
                    UINT32   array_size,
                    ADDR_48  addr_array[],
                    UINT32   count_array[],
                    UINT32  *p_addr_count)
   {
   LD_ELAN_CONTEXT   *p_elan;
   ARRAY_BLOCK        array_block;

   /* Obtain a pointer to the ELAN instance structure. */

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   /* Initialize the array block with a zero index, the maximum number of
    * entries to copy, and the pointers to the destination arrays.
    */
   array_block.addr_array  = addr_array;
   array_block.count_array = count_array;
   array_block.index       = 0;
   array_block.max         = array_size;

   /* Get the actual number of entries in the multicast table. */

   *p_addr_count = map_list_size_get (p_elan->mcast_list);

   /* Traverse the multicast list thus filling the target arrays. */

   map_list_traverse (p_elan->mcast_list,
                      mcast_get_handler,
                      (HANDLE) &array_block);

   return;
   }

/*++
* ===========
* = lec_xmt =
* ===========
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS lec_xmt (HANDLE   elan_handle,
                ADDR_48  dst_addr,
                void    *p_pkt,
                UINT32   length)
   {
   LD_ELAN_CONTEXT *p_elan;
   BOOLEAN          may_xmt;
   HANDLE           conn_handle;
   STATUS           status;
   UINT32           user_data;

   /* Get a pointer to the ELAN instance data block. */

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   /* If the ELAN is not up (i.e. this LEC is not fully joined to the ELAN),
    * return failure status.
    */
   if (p_elan->elan_up == FALSE)
      return STATUS_K_FAILURE;

   /* Attempt to resolve the MAC destination address to a Virtual
    * Connection.
    */
   lc_addr_resolve (p_elan->lc_elan_handle,
                    dst_addr,
                    &may_xmt,
                    &conn_handle);

   /* If the address resolution has allowed the packet to be transmitted to
    * the destination, send the packet using the provided connection handle.
    */
   if (may_xmt)
      {
      /* Set the user_data to indicate that this is a normal (protocol derived)
       * packet and that an LE header must be prepended.
       */
      user_data = USER_DATA_NORMAL_LE (p_elan->lec_id);

      EVENT (EM_XDATA, ("Data Packet Transmitted\n"));

      /* Instruct the Connection Manager to transmit the packet. and return
       * the status that was returned by the CM.
       */
      status    = cm_sap_xmt_vc (conn_handle, p_pkt, length, user_data, NULL);
      return status;
      }
   else
      {
      /* If the address resolution function denies permission to transmit,
       * return status indicating network congestion.
       */
      EVENT (EM_XDATA, ("Transmit Congestion per LEC_CTRL\n"));
      return STATUS_K_CONGESTED;
      }
   }


STATUS lec_attr_get (HANDLE     elan_handle,
                     UINT16     attr_id,
                     void      *p_out_buffer,
                     UINT16     out_buffer_length,
                     UINT16    *p_bytes_written)
   {
   LD_ELAN_CONTEXT *p_elan;
   UINT16           length;
   UINT32           value32;
   int              i;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   switch (attr_id)
      {
      case ATTR_C1_MY_ATM_ADDR:
      case ATTR_C2_LAN_TYPE:
      case ATTR_C3_LAN_MTU:
      case ATTR_C4_PROXY_FLAG:
      case ATTR_C5_ELAN_NAME:
      case ATTR_C6_LOCAL_ESI:
      case ATTR_C7_CONTROL_TIMEOUT:
      case ATTR_C8_LOCAL_RD:
      case ATTR_C9_LES_ATM_ADDR:
      case ATTR_C10_MAX_UF_COUNT:
      case ATTR_C11_MAX_UF_TIME:
      case ATTR_C12_VCC_TIMEOUT:
      case ATTR_C13_MAX_RETRY_COUNT:
      case ATTR_C14_LEC_ID:
      case ATTR_C16_LEARP_CACHE:
      case ATTR_C17_AGING_TIME:
      case ATTR_C18_FORW_DELAY_TIME:
      case ATTR_C19_TOPOLOGY_CHANGE_FLAG:
      case ATTR_C20_LEARP_RSP_TIME:
      case ATTR_C21_FLUSH_TIMEOUT:
      case ATTR_C22_PATH_SWITCH_DELAY:
      case ATTR_SM_STATE:
      case ATTR_SM_STATE_HISTORY:
      case ATTR_BUS_ATM_ADDR:
      case ATTR_ILLEGAL_CTRL_FRAMES_RVCD:
      case ATTR_CTRL_XMT_FAILURES:
      case ATTR_ILLEGAL_TRANSITIONS:
      case ATTR_CTRL_FRAMES_SENT:
      case ATTR_CTRL_FRAMES_RCVD:
      case ATTR_LEARPS_SENT:
      case ATTR_LEARPS_RCVD:
      case ATTR_UCASTS_SENT_DIRECT:
      case ATTR_UCASTS_FLOODED:
      case ATTR_UCASTS_DISCARDED:
      case ATTR_NUCASTS_SENT:
         return lc_attr_get (p_elan->lc_elan_handle,
                             attr_id,
                             p_out_buffer,
                             out_buffer_length,
                             p_bytes_written);

      case ATTR_BUS_FRAMES_RCVD:
         value32 = p_elan->bus_frames_rcvd;
         length  = 4;
         break;

      case ATTR_DIRECT_FRAMES_RCVD:
         value32 = p_elan->direct_frames_rcvd;
         length  = 4;
         break;

      case ATTR_RCV_FRAMES_DISCARDED:
         value32 = p_elan->rcv_frames_discarded;
         length  = 4;
         break;

      case ATTR_ECHOS_SUPPRESSED:
         value32 = p_elan->echos_suppressed;
         length  = 4;
         break;

      default:
         return STATUS_K_ARG_2_BAD;
      }

   if (out_buffer_length < length)
      return STATUS_K_ARG_4_BAD;

   *p_bytes_written = length;
   for (i = 0; i < length; i++)
      ((UINT8 *) p_out_buffer)[i] = ((UINT8 *) &value32)[i];
   return STATUS_K_SUCCESS;
   }

STATUS lec_attr_get_next (HANDLE     elan_handle,
                          UINT16     attr_id,
                          void      *p_in_buffer,
                          UINT16     in_buffer_length,
                          void      *p_out_buffer,
                          UINT16     out_buffer_length,
                          UINT16    *p_bytes_written)
   {
   LD_ELAN_CONTEXT *p_elan;
   UINT16           length;
   UINT32           value32;
   int              i;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   switch (attr_id)
      {
      case ATTR_C1_MY_ATM_ADDR:
      case ATTR_C2_LAN_TYPE:
      case ATTR_C3_LAN_MTU:
      case ATTR_C4_PROXY_FLAG:
      case ATTR_C5_ELAN_NAME:
      case ATTR_C6_LOCAL_ESI:
      case ATTR_C7_CONTROL_TIMEOUT:
      case ATTR_C8_LOCAL_RD:
      case ATTR_C9_LES_ATM_ADDR:
      case ATTR_C10_MAX_UF_COUNT:
      case ATTR_C11_MAX_UF_TIME:
      case ATTR_C12_VCC_TIMEOUT:
      case ATTR_C13_MAX_RETRY_COUNT:
      case ATTR_C14_LEC_ID:
      case ATTR_C16_LEARP_CACHE:
      case ATTR_C17_AGING_TIME:
      case ATTR_C18_FORW_DELAY_TIME:
      case ATTR_C19_TOPOLOGY_CHANGE_FLAG:
      case ATTR_C20_LEARP_RSP_TIME:
      case ATTR_C21_FLUSH_TIMEOUT:
      case ATTR_C22_PATH_SWITCH_DELAY:
      case ATTR_SM_STATE:
      case ATTR_SM_STATE_HISTORY:
      case ATTR_BUS_ATM_ADDR:
      case ATTR_ILLEGAL_CTRL_FRAMES_RVCD:
      case ATTR_CTRL_XMT_FAILURES:
      case ATTR_ILLEGAL_TRANSITIONS:
      case ATTR_CTRL_FRAMES_SENT:
      case ATTR_CTRL_FRAMES_RCVD:
      case ATTR_LEARPS_SENT:
      case ATTR_LEARPS_RCVD:
      case ATTR_UCASTS_SENT_DIRECT:
      case ATTR_UCASTS_FLOODED:
      case ATTR_UCASTS_DISCARDED:
      case ATTR_NUCASTS_SENT:
         return lc_attr_get_next (p_elan->lc_elan_handle,
                                  attr_id,
                                  p_in_buffer,
                                  in_buffer_length,
                                  p_out_buffer,
                                  out_buffer_length,
                                  p_bytes_written);

      default:
         return STATUS_K_ARG_2_BAD;
      }

   if (out_buffer_length < length)
      return STATUS_K_ARG_6_BAD;

   *p_bytes_written = length;
   for (i = 0; i < length; i++)
      ((UINT8 *) p_out_buffer)[i] = ((UINT8 *) &value32)[i];
   return STATUS_K_SUCCESS;
   }

STATUS lec_attr_set (HANDLE     elan_handle,
                     UINT16     attr_id,
                     void      *p_in_buffer,
                     UINT16     in_buffer_length)
   {
   LD_ELAN_CONTEXT *p_elan;

   p_elan = (LD_ELAN_CONTEXT *) elan_handle;

   switch (attr_id)
      {
      case ATTR_C1_MY_ATM_ADDR:
      case ATTR_C2_LAN_TYPE:
      case ATTR_C3_LAN_MTU:
      case ATTR_C4_PROXY_FLAG:
      case ATTR_C5_ELAN_NAME:
      case ATTR_C6_LOCAL_ESI:
      case ATTR_C7_CONTROL_TIMEOUT:
      case ATTR_C8_LOCAL_RD:
      case ATTR_C9_LES_ATM_ADDR:
      case ATTR_C10_MAX_UF_COUNT:
      case ATTR_C11_MAX_UF_TIME:
      case ATTR_C12_VCC_TIMEOUT:
      case ATTR_C13_MAX_RETRY_COUNT:
      case ATTR_C14_LEC_ID:
      case ATTR_C16_LEARP_CACHE:
      case ATTR_C17_AGING_TIME:
      case ATTR_C18_FORW_DELAY_TIME:
      case ATTR_C19_TOPOLOGY_CHANGE_FLAG:
      case ATTR_C20_LEARP_RSP_TIME:
      case ATTR_C21_FLUSH_TIMEOUT:
      case ATTR_C22_PATH_SWITCH_DELAY:
      case ATTR_SM_STATE:
      case ATTR_SM_STATE_HISTORY:
      case ATTR_BUS_ATM_ADDR:
      case ATTR_ILLEGAL_CTRL_FRAMES_RVCD:
      case ATTR_CTRL_XMT_FAILURES:
      case ATTR_ILLEGAL_TRANSITIONS:
      case ATTR_CTRL_FRAMES_SENT:
      case ATTR_CTRL_FRAMES_RCVD:
      case ATTR_LEARPS_SENT:
      case ATTR_LEARPS_RCVD:
      case ATTR_UCASTS_SENT_DIRECT:
      case ATTR_UCASTS_FLOODED:
      case ATTR_UCASTS_DISCARDED:
      case ATTR_NUCASTS_SENT:
      case ATTR_LEARP_SET:
      case ATTR_LEARP_DELETE:
         return lc_attr_set (p_elan->lc_elan_handle,
                             attr_id,
                             p_in_buffer,
                             in_buffer_length);

      case ATTR_BUS_FRAMES_RCVD:
      case ATTR_DIRECT_FRAMES_RCVD:
      case ATTR_RCV_FRAMES_DISCARDED:
      case ATTR_ECHOS_SUPPRESSED:
         /* Return failure because these are read-only attributes. */

         return STATUS_K_FAILURE;

      default:
         return STATUS_K_ARG_2_BAD;
      }
   }



