/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
 *
 * Module Name:
 *    svc.c
 *
 * Abstract:
 *    This file implements the interfaces between the connection manager and
 *    the signaling interface module.
 * 
 * Authors:
 *    JAD - John A. DeNisco
 *
 * Modification History:
 *    Date        Name   Description
 *    03-Feb-95   JAD    Created.
 *
 * Description:
 *    It is the primary responsibility of this module to make a clean interface
 *    between the connection manager, and a signaling interface module. This
 *    signaling interface module could be, for example a Network Call Coordinator,
 *    or a trillium signaling module interface. This module primarily translates
 *    data from svc structures to trillium as, or am structures.
 */

/* Header include files (.h)                                      */

/* Includes required by driver interface                          */

#include <stdio.h>
#include "g_types.h"       /* Driver type definitions             */
#include "codes.h"         /* Error codes                         */
#include "utl.h"           /* System utilities                    */
#include "utl_os.h"        /* Operating system specific utilities */

/* The following header files are required by the driver interface*/

#include "atm.h"           /* ATM definitions                     */
#include "system.h"        /* System definitions                  */
#include "svc_info.h"      /* Svc information structures          */
#include "svc.h"           /* General svc structures              */
#include "cm.h"            /* Connection manager interface        */
#include "cm_svc.h"        /* Connection manager to svc interface */
#include "cm_sap.h"        /* Connection manager/sap interface    */
#include "sig.h"           /* Signaling definitions               */
#include "g_event.h"       /* Event definitions                   */


/* Event constants */

#define EMOD  MOD_SIG
#define EINST p_sig_state->p_text
#define EMASK p_sig_state->event_mask

/* Local functions                                                */

PRIVATE void    svc_rcv_callback         (HANDLE            conn_context,
                                          HANDLE            conn_handle,
                                          UINT32            length,
                                          UINT32            user_data,
                                          AAL_DATA         *p_aal_data,
                                          void            **pp_packet);
PRIVATE void    svc_notify_callback      (HANDLE            conn_context,
                                          HANDLE            conn_handle,
                                          NOTIFY_EVENT      event,
                                          UINT16            reason,
                                          UINT32            endpoint_ref,
                                          BOOLEAN           calling_party);
PRIVATE void    svc_link_status_callback (HANDLE            conn_context,
                                          ATM_LINK_STATE    link_state);

typedef struct
   {
   HANDLE      gbl_os_handle;
   HANDLE      adap_os_handle;
   HANDLE      cm_handle;
   HANDLE      cm_sap_handle;
   HANDLE      conn_handle;
   } SVC_STATE;

/*++
* ==============
* = svc_create =
* ==============
*
* Description:
*  This function calls signaling module's create function (sig_create),
*  to initialize the signaling protocol stack. So the connection manager
*  can be binded with the svc module the cm_svc_config function is then
*  called. The svc module then registers with with the connection manager,
*  and then sets up the signaling pvc.
--*/

STATUS svc_create (HANDLE  gbl_os_handle,
                   HANDLE  adap_os_handle,
                   HANDLE  cm_handle,
                   char   *p_text,
                   HANDLE *p_svc_handle)
   {
   STATUS     status;
   QOS q;
   SVC_STATE   *p_state;

   p_state = (SVC_STATE *) os_mem_alloc (adap_os_handle,
                                         sizeof (SVC_STATE));
   if (p_state == NULL)
      return STATUS_K_RESOURCES;

   p_state->gbl_os_handle  = gbl_os_handle;
   p_state->adap_os_handle = adap_os_handle;
   p_state->cm_handle      = cm_handle;

   /* Initialize Signaling Stack (plus QSAAL) */

   /* Configure the connection manager with the svc handle. */

   cm_svc_config (cm_handle, (HANDLE) p_state);

   /* Register with the connection manager SAP */

   status = cm_sap_register(cm_handle,
                            *p_svc_handle,
                            svc_rcv_callback,
                            svc_notify_callback,
                            NULL,
                            svc_link_status_callback,
                            NULL,
                            p_text,
                           &(p_state->cm_sap_handle));
   if (status != STATUS_K_SUCCESS)
      return (status);

   /* Setup the PVC */

   status = cm_sap_pvc_setup(p_sig_state->cm_sap_handle,
                             *p_svc_handle,
                             ATMVPI_SIG,
                             ATMVCI_SIG,
                             &q,
                            &(p_state->conn_handle));
   if ((status != STATUS_K_SUCCESS) && (status != STATUS_K_PENDING))
      return (status);

   /* Return the status */

   *p_svc_handle = (HANDLE) p_state;
   return (status);
   }

/*++
* ===============
* = svc_destroy =
* ===============
*
* Description:
*  This function destroys the svc module. It first tears down the signaling
*  pvc. Then unregisters the connection manager sap, then calls sig_destroy
*  to destroy the signaling protocol stack.
--*/

void svc_destroy (HANDLE svc_handle)
   {
   SVC_STATE *p_state;

   /* Get the pointer to the state */

   p_state = (SVC_STATE *) svc_handle;
	if (p_state == NULL)
		return;

   /* Teardown the PVC */

   cm_sap_pvc_teardown(p_state->conn_handle);

   /* Unregister with the connection manager SAP */

   cm_sap_unregister(p_state->cm_sap_handle);

   /* Call the signaling destroy */

   os_mem_dealloc (p_state->adap_os_handle, p_state);
   }


/*++
* ======================
* = svc_event_mask_set =
* ======================
*
* Description:
*
*   Sets the event mask which defines which classes of events are to be
*   reported.
*
--*/                                              

void svc_event_mask_set (HANDLE svc_handle,
                         UINT32 event_mask)
   {
   SVC_STATE *p_state;

   /* Get the pointer to the state */

   p_state = (SVC_STATE *) svc_handle;
    
   /* Set the mask */

   p_state->event_mask = event_mask;
   }

/*++
* ===================
* = svc_cm_conn_req =
* ===================
*
* Description:
*  This function is a virtual connection request. For more information
*  see "cm_svc.h".
*
--*/

STATUS svc_cm_conn_req (HANDLE     svc_handle,
                        CALL_ID    call_user_id,
                        CONN_INFO *p_conn_info)
   {
   }

/*++
* ===================
* = svc_cm_conn_rsp =
* ===================
*
* Description:
*  This function is the virtual connection response. For more information
*  see "cm_svc.h".
*
--*/

STATUS svc_cm_conn_rsp (HANDLE       svc_handle,
                        CALL_ID      call_user_id,
                        CALL_ID      call_provider_id,
                        CONN_STATUS *p_conn_status)
   {
   }

/*++
* ========================
* = svc_cm_add_party_req =
* ========================
*
* Overview:
*  This function is the add party request.  For more information
*  see "cm_svc.h".
*
* Arguments:
*  svc_handle       - (IN) The handle to the instance to be used for this
*                          add party request.
*
*  call_user_id     - (IN) ID used to identify the user of this call instance.
*
*  call_provider_id - (IN) ID used to identify the provider of this call
*                          instance.
*
*  p_add_pty_info   - (IN) Pointer to the add party information.
*
* Returns:
*  STATUS_K_SUCCESS - Success, see postconditions.
*  STATUS_K_FAILURE - Not OK returned from Q93.B module.
*
* Preconditions:
*  None.
*
* Postconditions:
*  The add party message is transmitted.
--*/

STATUS svc_cm_add_party_req (HANDLE        svc_handle,
                             CALL_ID       call_user_id,
                             CALL_ID       call_provider_id,
                             ADD_PTY_INFO *p_add_pty_info)
   {
   }

/*++
* ========================
* = svc_cm_add_party_rsp =
* ========================
*
* Overview:
*  This function is an add party response. For more information see
*  "cm_svc.h".
*
--*/

STATUS svc_cm_add_party_rsp (HANDLE          svc_handle,
                             CALL_ID         call_user_id,
                             CALL_ID         call_provider_id,
                             ADD_PTY_STATUS *p_add_pty_status)
   {
   }


/*++
* ============================
* = svc_cm_add_party_rej_req =
* ============================
*
* Overview:
*  This function is the an add party reject request. For more information see
*  "cm_svc.h".
*
--*/

STATUS svc_cm_add_party_rej_req (HANDLE         svc_handle,
                                 CALL_ID        call_user_id,
                                 CALL_ID        call_provider_id,
                                 DROP_PTY_INFO *p_drop_pty_info)
   {
   }

/*++
* ========================
* = svc_cm_drop_party_req =
* ========================
*
* Overview:
*  This function is the drop party request. For more information see
*  "cm_svc.h".
*
--*/

STATUS svc_cm_drop_party_req (HANDLE        svc_handle,
                              CALL_ID       call_user_id,
                              CALL_ID       call_provider_id,
                              DROP_PTY_INFO *p_drop_pty_info)
   {
   }

/*++
* ==================
* = svc_cm_rel_req =
* ==================
*
* Overview:
*  This function is a release request. For more information see
*  "cm_svc.h".
--*/

STATUS svc_cm_rel_req (HANDLE        svc_handle,
                       CALL_ID       call_user_id,
                       CALL_ID       call_provider_id,
                       RELEASE_INFO *p_rel_cause)
   {
   }

/*++
* ======================
* = svc_cm_restart_req =
* ======================
*
* Overview:
*  This function is the restart request. For more information "see cm_svc.h".
*
--*/

STATUS svc_cm_restart_req (HANDLE svc_handle)
   {
   }

/*++
* ======================
* =  sap_rcv_callback  =
* ======================
*
* Overview:
*   Callback for received packets. A pointer to this routine is supplied
*   at registration time by the SAP.
*
* Arguments:
*   conn_context - (IN) svc_handle for the incoming VC.
*   conn_handle  - (IN) Handle of Connection on which the packet was received.
*   length       - (IN) Length of the packet payload in bytes.
*   user_data    - (IN) User data that was transferred from the layers below
*                       the CM to the SAP.
*   p_aal_data   - (IN) Data specific to the AAL being used is passed in this
*                       structure.
*   pp_packet    - (IN) Pointer to a pointer to a received packet.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The packet referred to by *pp_packet may be safely overwritten or
*   destroyed.
--*/

PRIVATE void svc_rcv_callback (HANDLE     conn_context,
                               HANDLE     conn_handle,
                               UINT32     length,
                               UINT32     user_data,
                               AAL_DATA  *p_aal_data,
                               void     **pp_packet)
   {
   SVC_STATE *p_state;

   /* Get the pointer to the state */

   p_state = (SVC_STATE *) svc_handle;
	if (p_state == NULL)
		return;

   /* Copy the entire frame from the receive buffer to local memory. */

   os_buff_hdr_copy (*pp_packet,
                     length,
                     p_mbuf->p_data,
                     length,
                     &bytes_copied);

   /* Call qsaal data indication function */

   }

/*++
* ==========================
* = svc_vc_notify_callback =
* ==========================
*
* Overview:
*   Callback to notify the SAP of the results of a VC setup.
*
* Arguments:
*   conn_context  - (IN) SVC handle.
*   conn_handle   - (IN) Not used.
*   event         - (IN) Type of event the SVC module is being notified of.
*   reason        - (IN) Not used.
*   endpoint_ref  - (IN) Not used.
*   calling_party - (IN) Not used.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
PRIVATE void svc_notify_callback (HANDLE         conn_context,
                                  HANDLE         conn_handle,
                                  NOTIFY_EVENT   event,
                                  UINT16         reason,
                                  UINT32         endpoint_ref,
                                  BOOLEAN        calling_party)
   {
   SVC_STATE *p_state;

   /* Get the pointer to the state */

   p_state = (SVC_STATE *) svc_handle;
	if (p_state == NULL)
		return;

   /* List the arguments not used */

   ARG_NOT_USED (conn_handle);
   ARG_NOT_USED (reason);
   ARG_NOT_USED (endpoint_ref);
   ARG_NOT_USED (calling_party);

   }

/*++
* ============================
* = svc_link_status_callback =
* ============================
*
* Overview:
*   The Connection Manager invokes this callback when the link status changes.
*   This serves to notify the SAP client when the link becomes available or
*   unavailable.
*
*   When the link becomes unavailable, the signaling link up indication will
*   called.
*
* Arguments:
*   conn_context - (IN) svc_handle.
*   link_status  - (IN) New value of link status.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
PRIVATE void svc_link_status_callback (HANDLE          conn_context,
                                       ATM_LINK_STATE  link_state)
   {
   SVC_STATE *p_state;

   /* Get the pointer to the state */

   p_state = (SVC_STATE *) svc_handle;
	if (p_state == NULL)
		return;

   /* Handle the event, we only handle line up and line down */

   switch (link_state)
      {
      case LINK_LINE_UP:
      /* Bring the signaling stack up */

         break;
      case LINK_DOWN:
      /* Bring the signaling stack down */

         break;
      }
   }

